
#include <fstream.h>

#include "proxy.h"
#include "strlib.h"

#define MAX_HEADER_LINE	16384
#define RELAY_BUF_SIZE	16384
#define MAX_METHOD	4
#define MAX_HOST	256
#define MAX_FILE	8192
#define MAX_PROTOCOL	16

static  char relay_buf[RELAY_BUF_SIZE];


void Proxy::fix_user_agent(char* relay_buf)
 {
  if(spoof_user_agent)
   if(!strncasecmp_to_delim(relay_buf, "User-agent", ':', RELAY_BUF_SIZE))
    {
     sprintf(relay_buf, "User-agent: %s\r\n", spoof_user_agent);
    }
 }

void Proxy::handle_client(int client_socket_fd)
 {
  char header_buf[MAX_HEADER_LINE];
  
  FILE *in  = fdopen(client_socket_fd, "r");
  FILE *out  = fdopen(client_socket_fd, "w");

  setbuf(in, NULL);
  setbuf(out, NULL);

  if(!in)
   {
    warn( "Proxy:fdopen failed on 'in': %s\n", strerror(errno));
    return;
   }

  if(!out)
   {
    warn( "fdopen failed on 'out': %s\n", strerror(errno));
    return;
   }

  char method[MAX_METHOD + 1];
  char host[MAX_HOST + 1];
  int port;
  char file[MAX_FILE + 1];
  char protocol[MAX_PROTOCOL + 1];
 
  int num_bytes;

  if(!fgets(header_buf, sizeof(header_buf), in))
   {
    warn("Error processing HTTP request");
    exit(1);
   }

  int header_buf_len = strlen(header_buf);
  int save_c = header_buf[header_buf_len - 1];
  header_buf[header_buf_len - 1]  = 0;
  info("%s %s", get_peer_addr(client_socket_fd), header_buf);
  header_buf[header_buf_len - 1]  = save_c;
 
  parse_request(header_buf, method, host, port, file, protocol);
  sprintf(header_buf, "%s %s %s\r\n", method, file, protocol);
  
  TcpClient* http_client = NULL;

    http_client = new TcpClient(host, port, log);
    http_client->connect();

   if(!*http_client)
    {
     warn("Aborted HTTP client");
     exit(0);
    }

   if((num_bytes = http_client->send(header_buf, strlen(header_buf))) == -1)
    {
     warn("Error sending the header to the server: %s\n", strerror(errno));
     return;
    }

  debug(4, "Wrote %d bytes of the first line of the header",
    num_bytes);

  if(strcasecmp("GET", method) == 0)
   send_get(in, out, http_client);

  if(strcasecmp("POST", method) == 0)
   send_post(in, out, http_client);
 
  forward_reply(in, out, http_client);
 }

void Proxy::send_post(FILE* in, FILE* out, TcpClient* http_client)
 {
  int num_bytes;
  int content_len = -1;

  while(1)
   {
      if(!fgets(relay_buf, sizeof(relay_buf), in))
       {
        warn("Unexpected termination of HTTP header");
	exit(0);
       }

      fix_user_agent(relay_buf);
       
      num_bytes = strlen(relay_buf);
      
      //fprintf(stderr, "num_bytes = %d\n", num_bytes);
      const char* content_len_str = "Content-length";
      char token_buf[sizeof(relay_buf)];

      int i;
      
      if(content_len < 0) 
       {
        token_buf[0] = 0;

	for(i = 0; i < sizeof(token_buf); i++)
         {
          if(isspace(relay_buf[i]) || relay_buf[i] == ':')
	   {
	    token_buf[i] = 0;
	    break;
      	   }
	  else
	    token_buf[i] = relay_buf[i];
	  }
	
	 if(strcasecmp(token_buf, content_len_str) == 0)
           { 
	     for(; i < sizeof(relay_buf); i++)
	       if(relay_buf[i] == ':') 
	         {
		  i++;
		  break;
		 }

 	      for(; i < sizeof(relay_buf); i++)
	       if(isdigit(relay_buf[i])) 
	         break;
 
	       content_len = 0;
	       //fprintf(stderr, "i = %d relay_buf +i = '%s'\n",
	       //  i, relay_buf + i);

 	     for(; i < sizeof(relay_buf); i++)
	      {
	       if(!isdigit(relay_buf[i])) 
	         break;
	       else
	        content_len = 10 * content_len + relay_buf[i] - '0'; 
	      }
           
	      //fprintf(stderr, "computed content_len = %d\n", content_len);
	    }
      	 }
       
     
      //fprintf(stderr, "final content_len = %d\n", content_len);
      if(relay_buf[0] == '\n' || relay_buf[0] == '\r' )
       {
        http_client->send("\r\n", 2);
        
	if(content_len > 0)
	 {
	  char* content_buf = new char[content_len];
	  while(content_len > 0)
	   {
	    num_bytes = fread(content_buf, 1, content_len, in);
	    char info_fmt_buf[64];
	    snprintf(info_fmt_buf, sizeof(info_fmt_buf), "POST content: %%%ds",
	      content_len);
	    info(info_fmt_buf, content_buf);
	    if(num_bytes <= 0)
	     {
	      warn("Error in POST request");
	      exit(1);
	     }
	    
	    http_client->send(content_buf, num_bytes);
	    content_len -= num_bytes;
	   }

	  delete[] content_buf;
	 }
	//fprintf(stderr, "End of header\n");
	break;
       }
    
      if( http_client->send(relay_buf, num_bytes) == -1)
        warn("Error sending the data: %s\n", strerror(errno));
   }

 }

void Proxy::forward_reply(FILE* in, FILE* out, TcpClient *http_client)
 {
  pid_t filter_pid;
  int pipe_to_filter[2] = {-1, -1};
  int pipe_from_filter[2] = {-1, -1};
  // if we are filtering, we need two pipes, one to send the data to the filter app
  // and the other to receive the filtered data back

  if(filter_app)
   {
    debug(4, "Using filter_app");
    pipe_on = 1;

     if(pipe(pipe_to_filter) == -1)
      {
       warn("Could not open pipe for %s, will not filter", filter_app);
       pipe_on = 0;
      }
  
    if(pipe_on)
     if(pipe(pipe_from_filter) == -1)
     {
      warn("Could not open pipe for %s, will not filter", filter_app);
      pipe_on = 0;
     }
   
    if(pipe_on)
     switch(filter_pid = fork())
      {
       case 0:
        if(close(pipe_from_filter[0]) == -1) // the child will not read from the _from_ pipe
	 {
	  warn("could not close pipe file descritor, will not filter");
	  pipe_on = 0;
	  break;
	 } 
	
	if(dup2(pipe_from_filter[1], 1) == -1) // redirect stdout to the pipe
	  {
	   warn("dup2() failed, will not filter");
	   pipe_on = 0;
	   break;
	  }
       
        close(pipe_from_filter[1]);

        if(close(pipe_to_filter[1]) == -1) // the child will not write to the _to_ pipe
	 {
	  warn("could not close pipe file descritor, will not filter");
	  pipe_on = 0;
	  break;
	 } 
	
	if(dup2(pipe_to_filter[0], 0) == -1) // redirect stdin to the pipe
	  {
	   warn("dup2() failed, will not filter");
	   pipe_on = 0;
	   break;
	  }
       
        close(pipe_to_filter[0]);
	
	if(execv(filter_bin, filter_args) == -1) 
	   {
	    warn("execv(%s, %p) failed in child: ", filter_bin, filter_args);
	    exit(1);
	   }
	   break;
      case -1:
        warn("Could not fork() for %s, will not filter", filter_app);
        pipe_on = 0;
	break; 
      }
   
    if(!pipe_on)
     {
      close(pipe_to_filter[0]);
      close(pipe_from_filter[0]);
      close(pipe_to_filter[1]);
      close(pipe_from_filter[1]);
     }
    else
     {
      close(pipe_to_filter[0]);  // parent will not read from the _to_ pipe
      close(pipe_from_filter[1]);  // parent will not write to the  _from_ pipe
      //waitpid(filter_pid, 0, 0);
     }
   }
  else
   debug(4, "using no filter");
  //char pipe_buf[sizeof(relay_buf)];

  int num_bytes;

  while((num_bytes = http_client->receive(relay_buf, sizeof(relay_buf))) > 0)
   {
    if(pipe_on)
     {
      int pipe_bytes;
      pipe_bytes = write(pipe_to_filter[1], relay_buf, num_bytes) ;
      if(pipe_bytes == -1)
       {
        error("write() to filter pipe failed, exiting the thread");
	exit(1);
       }

      if(pipe_bytes < num_bytes)
       {
        error(
	 "full write() to pipe failed, exiting thread.\
	   Wait until the next release to deal with this properly");
        exit(1);
       }

      debug(4, "wrote %d bytes into the filter pipe", pipe_bytes);
      
      // now this stuff is to keep the buffer pipe from overflowing
      // the filter will try to write at some point, and if we do not read it
      // the data will get stuck in the pipe buffer
      fd_set pipe_fds;
      struct timeval pipe_tv;
      pipe_tv.tv_sec = 0;
      pipe_tv.tv_usec = 300;
      FD_ZERO(&pipe_fds);
      FD_SET(pipe_from_filter[0], &pipe_fds);

      int filter_bytes;
      int select_ret = select(pipe_from_filter[0] + 1, &pipe_fds, NULL, NULL,
       &pipe_tv);

      switch(select_ret)
       {
        case 0: 
	  debug(4, "select() timed out, filter does not yet have the data");
	  break;
	case -1:
	 debug(4, "error on select() reading from the filter: %s", 
	   strerror(errno)); 
	  break;

	default:
	 
	 debug(4, "filter has the data after select()");
	 filter_bytes = read(pipe_from_filter[0], relay_buf, sizeof(relay_buf));
	 debug(4, "read %d bytes from the filter", filter_bytes);
         fwrite(relay_buf, 1, filter_bytes, out);
	 break;
       }
       
     }
    else
     {
      fwrite(relay_buf, 1, num_bytes, out);
     }
   }
 
  if(pipe_on)
   {
    close(pipe_to_filter[1]);
    while((num_bytes = read(pipe_from_filter[0], relay_buf, sizeof(relay_buf))) > 0)
      fwrite(relay_buf, 1, num_bytes, out);
   
    close(pipe_from_filter[0]);
   }

  pipe_on = 0;
 
 }

void Proxy::send_get(FILE* in, FILE* out, TcpClient* http_client)
 {
  int num_bytes;


  while(1)
   {
      if(!fgets(relay_buf, sizeof(relay_buf), in))
       {
        warn("Unexpected termination of GET HTTP header");
	exit(0);
       }
    
      fix_user_agent(relay_buf);
      num_bytes = strlen(relay_buf);
      debug(4, "GET header is %d bytes long", num_bytes);

      if(relay_buf[0] == '\n' || relay_buf[0] == '\r' )
       {
        http_client->send("\r\n", 2);
        debug(4, "End of GET header");
	break;
       }
    
      if( http_client->send(relay_buf, num_bytes) == -1)
        warn("Error sending the data: %s\n", strerror(errno));
   }

 }


void Proxy::parse_request(char *request, char* method, char* host,
   int& port,  char* file,
       char* protocol)
  {
   int j = 0;
   int i = 0;

   for(i = 0; request[j] != ' ' && i < MAX_METHOD; i++,j++)
    {
     method[i]  = request[j];
    }

   method[i] = 0;
   i =  0;

   for(; request[j] != '/' && j < MAX_HEADER_LINE; j++);
     
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }

   if(request[++j] != '/')
    {
     warn("request syntax error in request, '/' expected");
     exit(1);
    }

   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }

   j++;

   for(i = 0; i < MAX_HOST && request[j] != '/' && j < MAX_HEADER_LINE; i++,j++)
    {
     host[i] = request[j];
    }
 
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }
 
   if(i == MAX_HOST)
    {
     warn("host string too long");
     exit(1);
    }

   host[i] = 0;

   for(i = 0; i < MAX_FILE && request[j] != ' ' && j < MAX_HEADER_LINE; i++,j++)
    {
     file[i] = request[j];
    }
 
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }
 
   if(i == MAX_FILE)
    {
     warn("file string too long");
     exit(1);
    }

   file[i] = 0;
   j++;

   for(i = 0; i < MAX_PROTOCOL && 
     (request[j] != '\n' && request[j] != '\r') && j < MAX_HEADER_LINE; i++,j++)
    {
     protocol[i] = request[j];
    }
 
   if(j == MAX_HEADER_LINE)
    {
     warn("request header line too long");
     exit(1);
    }
 
   if(i == MAX_PROTOCOL)
    {
     warn("protocol string too long");
     exit(1);
    }

    protocol[i] = 0;

    char* str = strtok(host, ":");
    str = strtok(NULL, ":");
    if(str)
     {
      debug(4, "str = %s\n", str);
      port = atoi(str);
      *(--str) = 0;
     }
    else
     port = 80;
  }

#ifdef DEBUG
int main(int c, char** args)
 {
    Proxy p(8000);
    int i;
    char* requests[] = {
     "GET http://www.direct1.com/cgi-bin/survey?hello=world HTTP/1.0\n",
     "GET http://www.faneuil.com:8000/cgi-bin/survey?hello=world HTTP/1.0\n",
       NULL};

    for(i = 0; requests[i]; i++)
     {
      char method[MAX_METHOD + 1];
      char host[MAX_HOST + 1];
      char file[MAX_FILE + 1];
      char protocol[MAX_PROTOCOL + 1];
      int port;

      //fprintf(stderr, "parsing %s\n", requests[i]);
      p.parse_request(requests[i], method, host, port, file, protocol);
      fprintf(stderr, "got method = '%s', host = '%s', port = %d, file = '%s', protocol = '%s'\n",
        method, host, port, file, protocol);
     }
 }
#endif



