#include "proxy.h"
#include "strlib.h"
 
void die(char* fmt, ...);

void die(char* fmt, ...)
  {
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);

    exit(1);
  }


int main(int c, char** argv)
 {
  int port = 8000;
  int debug_level = 0;
  char* config_file = NULL;
  int listen_queue_size = 50;
  char log_file[256] = {"www_proxy.log"};
  char spoof_user_agent[1024] = {'\0'};
  char filter_app[1024] = {'\0'};

  if(c != 2) 
   die("Usage: www_proxy config_file\n");
  
  config_file = argv[1];

  ifstream is(config_file);
  
  if(!is)
    die("could not open %s: %s\n", config_file, strerror(errno));
 
  int line_num = 0;

  while(1)
      {
       char line[512];
       is.getline(line, sizeof(line));
       if(is.eof())
        break;
       line_num++;
      
       if(line[0] == '#')
        continue;

       char* param_name = strtok(line, " ");

       if(!param_name)
        continue;
   
       char* param_val = NULL;
       

       if(!strcmp("port", param_name))
        {
         param_val = strtok(NULL, " ");
	 if(param_val)
	  port = atoi(param_val);
	 else
	  die("Syntax error on line %d: missing port value\n", line_num);
	}
       else if(!strcmp("debug_level", param_name))
        {
         param_val = strtok(NULL, " ");
	 if(param_val)
	  debug_level = atoi(param_val);
	 else
	  die("Syntax error on line %d: missing debug_level value\n", line_num);
	}
       else if(!strcmp("spoof_user_agent", param_name))
        {
         param_val = strtok(NULL, "\"");
	 
	 if(param_val)
	  strn_delim_cpy(spoof_user_agent ,param_val,
	    sizeof(spoof_user_agent), '"');
	 else
	  die("Syntax error on line %d: missing spoof_user_agent value\n", line_num);
	 //fprintf(stderr, "Spoofing user agent: value set to %s\n", spoof_user_agent);
	}
       else if(!strcmp("filter_app", param_name))
        {
         param_val = strtok(NULL, "\"");
	 
	 if(param_val)
	  strn_delim_cpy(filter_app ,param_val,
	    sizeof(filter_app), '"');
	 else
	  die("Syntax error on line %d: missing filter_app value\n", line_num);
	 
	 //fprintf(stderr, "filter_app: value set to %s\n", filter_app);
	}
       else if(!strcmp("log_file", param_name))
        {
         param_val = strtok(NULL, "\"");

	 if(param_val)
	  strn_delim_cpy(log_file ,param_val, sizeof(log_file), '"');
	 else
	  die("Syntax error on line %d: missing log_file value\n", line_num);
	}
       else
	  die("Syntax error on line %d: %s -invalid parameter name\n",
	    line_num, param_name);

      }


   Proxy p(port, listen_queue_size, log_file);
   p.set_debug_level(debug_level);
   
   if(spoof_user_agent[0])
    p.set_spoof_user_agent(spoof_user_agent);
 
   if(filter_app[0])
    p.set_filter_app(filter_app);


   p.run_server();
 }

